use galactica_content::SpriteHandle;
use galactica_util::to_radians;
use nalgebra::{Point2, Vector2};

use super::{SpriteRect, UiElement};
use crate::{vertexbuffer::types::UiInstance, PositionAnchor, RenderInput, RenderState};

pub struct UiSprite {
	sprite: SpriteHandle,
	mask: Option<SpriteHandle>,

	rect: SpriteRect,
	children_under: Vec<Box<dyn UiElement>>,
	children_above: Vec<Box<dyn UiElement>>,
}

impl UiSprite {
	pub fn new(sprite: SpriteHandle, mask: Option<SpriteHandle>, rect: SpriteRect) -> Self {
		return Self {
			sprite,
			mask,
			rect,
			children_under: Vec::new(),
			children_above: Vec::new(),
		};
	}

	/// Add a child under this sprite
	pub fn add_child_under(&mut self, child: Box<dyn UiElement>) {
		self.children_under.push(child);
	}

	/// Add a child above this sprite
	//pub fn add_child_above(&mut self, child: Box<dyn UiElement>) {
	//	self.children_above.push(child);
	//}

	/// Add this image to the gpu sprite buffer
	pub fn push_to_buffer(&self, input: &RenderInput, state: &mut RenderState) {
		let pos = Point2::new(self.rect.pos.x, self.rect.pos.y);
		let dim = Vector2::new(
			self.rect.dim.y * input.ct.get_sprite(self.sprite).aspect,
			self.rect.dim.y,
		);

		for c in &self.children_under {
			c.push_to_buffer_child(input, state, pos, dim);
		}

		let sprite = input.ct.get_sprite(self.sprite);
		let texture_a = sprite.get_first_frame(); // ANIMATE

		state.push_ui_buffer(UiInstance {
			anchor: PositionAnchor::CC.to_int(),
			position: pos.into(),
			angle: to_radians(90.0),
			size: dim.y,
			color: [1.0, 1.0, 1.0, 1.0],
			texture_index: [texture_a, texture_a],
			texture_fade: 1.0,
			mask_index: self
				.mask
				.map(|x| {
					let sprite = input.ct.get_sprite(x);
					let texture_b = sprite.get_first_frame(); // ANIMATE
					[1, texture_b]
				})
				.unwrap_or([0, 0]),
		});

		for c in &self.children_above {
			c.push_to_buffer_child(input, state, pos, dim);
		}
	}
}

impl UiElement for UiSprite {
	/// Add this image to the gpu sprite buffer,
	/// as a child of another sprite
	fn push_to_buffer_child(
		&self,
		input: &RenderInput,
		state: &mut RenderState,
		parent_pos: Point2<f32>,
		parent_size: Vector2<f32>,
	) {
		let zero = Point2::new(
			parent_pos.x - (parent_size.x / 2.0),
			parent_pos.y + (parent_size.y / 2.0),
		);

		let pos = zero
			+ Vector2::new(
				self.rect.pos.x * parent_size.x,
				-self.rect.pos.y * parent_size.y,
			);
		let dim = Vector2::new(
			self.rect.dim.x * parent_size.x,
			self.rect.dim.y * parent_size.y,
		);

		for c in &self.children_under {
			c.push_to_buffer_child(input, state, pos, dim);
		}

		let sprite = input.ct.get_sprite(self.sprite);
		let texture_a = sprite.get_first_frame(); // ANIMATE

		state.push_ui_buffer(UiInstance {
			anchor: PositionAnchor::CNw.to_int(),
			position: pos.into(),
			angle: to_radians(90.0),
			size: dim.y,
			color: [1.0, 1.0, 1.0, 1.0],
			texture_index: [texture_a, texture_a],
			texture_fade: 1.0,
			mask_index: self
				.mask
				.map(|x| {
					let sprite = input.ct.get_sprite(x);
					let texture_b = sprite.get_first_frame(); // ANIMATE
					[1, texture_b]
				})
				.unwrap_or([0, 0]),
		});

		for c in &self.children_above {
			c.push_to_buffer_child(input, state, pos, dim);
		}
	}
}
